#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2008-2009, Sebastian Ramacher
# This file is released under a modified BSD license. See LICENSE.txt for
# details.
################################################################################

from btdt.secure_server import SecureXMLRPCServer
from btdt.common import AuthorizedXMLRPCRequestHandler, password_hash, \
                        AuthorizedServerProxy, random_string
from btdt.database import User, Session
from btdt.logger import logger
from thread import interrupt_main
from threading import Event

import socket

class AdminRPC(SecureXMLRPCServer):
    allow_reuse_address = 1
    functions = list()

    def register(method, functions=functions):
        functions.append(method.__name__)
        return method

    def is_allowed(self, user, pw):
        if user == self.__quit_user and pw == password_hash(self.__quit_pw):
            return True

        session = Session()
        return session.query(User).filter_by(name=user, password_hash=pw,
                                             is_admin=True).count() == 1

    def __init__(self, host_port, tracker, rpc, key_file, cert_file,
                 serve=True):
        logger().info('initializing admin XML-RPC server ...')

        self.doneevent = Event()
        self.host_port = host_port
        self.tracker = tracker
        self.rpc = rpc

        self.__quit_user = random_string(64)
        self.__quit_pw = random_string(64)

        if serve:
            SecureXMLRPCServer.__init__(
                self, host_port, key_file, cert_file,
                logRequests=False,
                HandlerClass=AuthorizedXMLRPCRequestHandler
            )

            for name in self.functions:
                self.register_function(getattr(self, name))

            session = Session()
            if not session.query(User).filter_by(is_admin=True).count():
                logger().warn("No admin user found in database! I'll create a "
                              "default admin ... please create a new admin and "
                              "delete user 'admin' afterwards!")
                self.add_user('admin', 'admin_password')
                self.add_admin('admin')

    def serve_forever(self):
        logger().info('admin XML-RPC server starts serving ...')
        while not self.doneevent.isSet():
            try:
                self.handle_request()
            except socket.error, e:
                logger().error('encountered some socket error: %s' % str(e))

    def shutdown(self):
        logger().info('shutting down admin XML-RPC server ...')

        host, port = self.host_port
        if not host:
            host = 'localhost'
        p = AuthorizedServerProxy('https://%s:%d' % (host, port),
                                  self.__quit_user, self.__quit_pw)
        try:
            p.quit()
        except socket.error, e:
            logger().info('admin XML-RPC server shut down caused a socket ' \
                          'error: %s' % str(e))

        logger().info('admin XML-RPC server shut down done')

    @register
    def quit(self):
        self.doneevent.set()
        # just handle the request and do nothing
        return True

    @register
    def list_files(self):
        return self.rpc.list_files()

    @register
    def delete_torrent(self, torrent):
        return False

    @register
    def update_torrents(self):
        self.tracker.force_parse_allowed()
        return self.list_files()

    @register
    def update_config(self, **kwargs):
        pass

    @register
    def add_user(self, user, pw, pre_hashed=False):
        if not pre_hashed:
            pw = password_hash(pw)

        session = Session()
        q = session.query(User).filter_by(name=user)
        if q.count():
            u = q.one()
            u.password_hash = pw
        else:
            u = User(user, pw)
            session.add(u)
        session.flush()

        return True

    @register
    def delete_user(self, user):
        session = Session()
        try:
            u = session.query(User).filter_by(name=user).one()
        except:
            raise ValueError("no user of name '%s'" % user)

        session.delete(u)
        session.flush()
        return True

    @register
    def list_users(self):
        session = Session()
        return [u.name for u in session.query(User)]

    @register
    def list_admins(self):
        session = Session()
        return [u.name for u in
                session.query(User).filter_by(is_admin=True)]

    @register
    def add_admin(self, user):
        session = Session()
        try:
            u = session.query(User).filter_by(name=user).one()
        except:
            raise ValueError("no user of name '%s'" % user)

        u.is_admin = True
        session.flush()

        return True

    @register
    def delete_admin(self, user):
        session = Session()
        try:
            u = session.query(User).filter_by(name=user).one()
        except:
            raise ValueError("no user of name '%s'" % user)

        u.is_admin = False
        session.flush()

        return True

    @register
    def shutdown_server(self):
        interrupt_main()
        return True
