#!/usr/bin/env python
# -*- coding: utf-8 -*-
# $Id$
# Copyright (C) 2008, Sebastian Ramacher
# Copyright (C) 2008, Andreas Stuehrk
# This file is released under a modified BSD license. See LICENSE.txt for
# details.
################################################################################

""" Secure HTTP and XML-RPC server

Provides a secure XML-RPC server and a HTTPS server based on
http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/496786 and
http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/442473.
It also provides functions to create self-signed certificates.
"""

from __future__ import with_statement

from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler
from SimpleXMLRPCServer import SimpleXMLRPCDispatcher
from SimpleXMLRPCServer import SimpleXMLRPCRequestHandler
import random
import socket

class SecureHandlerBase(object):
    """ Secure request handler base

    Provides basic functionality common to secure request handlers.
    """

    def setup(self):
        self.connection = self.request
        self.rfile = socket._fileobject(self.request, "rb", self.rbufsize)
        self.wfile = socket._fileobject(self.request, "wb", self.wbufsize)

class SecureHTTPRequestHandler(SecureHandlerBase, BaseHTTPRequestHandler):
    """ Secure HTTP request handler

    Provides basic functionality to handle HTTPS requests.
    """

    def __init__(self, *args, **kwargs):
        BaseHTTPRequestHandler.__init__(self, *args, **kwargs)

class SecureHTTPServer(HTTPServer):
    """ Secure HTTP server

    A HTTP server using HTTPS connections.
    """

    def __init__(self, server_address, key_file, cert_file,
                 HandlerClass=SecureHTTPRequestHandler):
        """
        Takes the same arguments as HTTPServer.__init__ with the exception
        of key_file as the second and cert_file as the third argument.
        """

        from OpenSSL import SSL

        HTTPServer.__init__(self, server_address, HandlerClass)
        ctx = SSL.Context(SSL.SSLv23_METHOD)
        ctx.use_privatekey_file(key_file)
        ctx.use_certificate_file(cert_file)
        self.socket = SSL.Connection(ctx, socket.socket(self.address_family,
                                                        self.socket_type))
        self.server_bind()
        self.server_activate()

class SecureXMLRPCRequestHandler(SecureHandlerBase, SimpleXMLRPCRequestHandler):
    """ Secure XML-RPC request handler

    It it very similar to SimpleXMLRPCRequestHandler but it uses HTTPS for
    transporting XML data.
    """

    def __init__(self, *args, **kwargs):
        """
        Takes the same arguments as SimpleXMLRPCRequestHandler.__init__.
        """

        SimpleXMLRPCRequestHandler.__init__(self, *args, **kwargs)

    def do_POST(self):
        """ Handles the HTTPS POST request

        It was copied out from SimpleXMLRPCServer.py and modified to shutdown
        the socket cleanly.
        """

        try:
            # get arguments
            data = self.rfile.read(int(self.headers["content-length"]))
            # In previous versions of SimpleXMLRPCServer, _dispatch
            # could be overridden in this class, instead of in
            # SimpleXMLRPCDispatcher. To maintain backwards compatibility,
            # check to see if a subclass implements _dispatch and dispatch
            # using that method if present.
            response = self.server._marshaled_dispatch(
                    data, getattr(self, '_dispatch', None)
                )
        except: # This should only happen if the module is buggy
            # internal error, report as HTTP server error
            self.send_response(500)
            self.end_headers()
        else:
            # got a valid XML RPC response
            self.send_response(200)
            self.send_header("Content-Type", "text/xml")
            self.send_header("Content-Length", str(len(response)))
            self.end_headers()
            self.wfile.write(response)

            # shut down the connection
            self.wfile.flush()
            self.connection.shutdown() # Modified here!

class SecureXMLRPCServer(SecureHTTPServer, SimpleXMLRPCDispatcher):
    """Secure XML-RPC server.

    It is very similar to SimpleXMLRPCServer but it uses HTTPS for transporting
    XML data.
    """

    def __init__(self, server_address, key_file, cert_file,
                 HandlerClass=SecureXMLRPCRequestHandler, allow_none=False,
                 encoding=None, logRequests=True):
        """
        Takes the same arguments as SimpleXMLRPCServer.__init__ with the
        exception of key_file as the second and cert_file as the third
        argument.
        """

        self.logRequests = logRequests

        try:
            SimpleXMLRPCDispatcher.__init__(self)
        except TypeError:
            # An exception is raised in Python 2.5 as the prototype of the
            # __init__ method has changed and now has 3 arguments (self,
            # allow_none, encoding)
            SimpleXMLRPCDispatcher.__init__(self, allow_none, encoding)

        SecureHTTPServer.__init__(self, server_address, key_file, cert_file,
                                  HandlerClass)

def openssl_version():
    """
    Retruns the version as a tuple of ints with (major, minor).
    """

    import OpenSSL
    return tuple(map(int, OpenSSL.__version__.split('.')))

def create_certificate(host_name, key_file, cert_file, key_type=None,
                       bits=4096, digest="sha1", serial=-1, not_before=0,
                       not_after=60*60*24*365*5):
    """
    Creates a key pair and self-signed certificate and saves them to
    key_file and cert_file, respectively.
    """

    # check OpenSSL version
    assert openssl_version() >= (0, 6), 'btdt requires at least pyOpenSSL 0.6'
    from OpenSSL import crypto

    if key_type is None:
        key_type = crypto.TYPE_RSA
    if serial == -1:
        serial = random.randint(0, 2**31)

    # create public/private key pair
    pkey = crypto.PKey()
    pkey.generate_key(key_type, bits)

    # create a certificate request
    req = crypto.X509Req()
    subj = req.get_subject()
    subj.CN = host_name

    req.set_pubkey(pkey)
    req.sign(pkey, digest)

    # create the certificate
    cert = crypto.X509()
    cert.set_serial_number(serial)
    cert.set_issuer(req.get_subject())
    cert.set_subject(req.get_subject())
    cert.set_pubkey(req.get_pubkey())
    cert.gmtime_adj_notBefore(not_before)
    cert.gmtime_adj_notAfter(not_after)
    cert.sign(pkey, digest)

    with open(key_file, 'w') as f:
        f.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, pkey))
    with open(cert_file, 'w') as f:
        f.write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert))

    return (pkey, cert)
