#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2008, Andreas Stuehrk
# Copyright (C) 2008, Sebastian Ramacher
# This file is released under a modified BSD license. See LICENSE.txt for
# details.
################################################################################

""" Webseeder

Provides a webseed server and a related class to throttle the upload rate.
"""

from __future__ import with_statement

import os
import socket
import time

from BaseHTTPServer import HTTPServer
from btdt.common import flatten, furlparse
from btdt.logger import logger
from httplib import HTTPConnection
from operator import itemgetter
from Queue import Queue
from SimpleHTTPServer import SimpleHTTPRequestHandler
from threading import Event, RLock, Semaphore, Thread
from urllib import unquote

class RateLimiter(object):
    def __init__(self, rate_limit):
        self.rate_limit = float(rate_limit)
        self.last_time = time.time()
        self.loaders = 0
        self._lock = RLock()

    def __call__(self, bytes):
        if self.rate_limit < 0:
            return
        self._lock.acquire()
        elapsed_time = time.time() - self.last_time
        if elapsed_time:
            expected_time = bytes / (self.rate_limit / self.loaders)
            wait_time = expected_time - elapsed_time
            self._lock.release()
            if wait_time > 0:
                time.sleep(wait_time)
            self.last_time = time.time()
        else:
            self._lock.release()

    def __enter__(self):
        self._lock.acquire()
        self.loaders += 1
        self._lock.release()
        return self

    def __exit__(self, type, value, traceback):
        self._lock.acquire()
        self.loaders -= 1
        self._lock.release()
        return False

class WebseedRequestHandler(SimpleHTTPRequestHandler):
    """ Webseed request handler """

    protocol_version = 'HTTP/1.1'

    def log_message(self, fmt, *args):
        logger().debug('%s: %s' % (self.client_address[0], fmt % args))

    def send_headers(self, headers, end_headers=False):
        for key, value in headers.iteritems():
            self.send_header(key, value)
        if end_headers:
            self.end_headers()

    def send(self, code, headers, data):
        try:
            if not 'Content-Length' in headers:
                headers['Content-Length'] = len(data)

            self.send_response(code)
            self.send_headers(headers, True)
            self.wfile.write(data)
        except socket.error:
            pass

    def send_file(self, code, headers, path, start, bytes=None):
        assert start >= 0
        try:
            self.send_response(code)
            if not 'Content-Length' in headers:
                headers['Content-Length'] = bytes
            self.send_headers(headers, True)
            file_size = os.stat(path).st_size
            if bytes is None:
                bytes = start - file_size
            assert (start + bytes) <= file_size
            with open(path, 'rb') as fo:
                if start:
                    fo.seek(start)
                with self.server.upload_limiter:
                    while bytes:
                        to_read = min(bytes, 8192)
                        self.wfile.write(fo.read(8192))
                        self.server.upload_limiter(to_read)
                        bytes -= to_read
        except socket.error:
            pass

    def do_QUIT(self):
        """ Handles QUIT request and shuts down server """
        self.server.doneevent.set()
        self.server.request_queue.put((None, None))
        return self.send(200, {}, '')

    def do_GET(self):
        try:
            (scheme, netloc, path, pars, query, fragment) = furlparse(self.path)
            path = unquote(path)[1:]
            # Newer versions of libtorrent use utf8 to encode the path.
            # Therefore we try to decode path.
            try:
                path = path.decode('utf8')
            except UnicodeDecodeError:
                pass

            params = {}
            for s in query.split('&'):
                if s != '':
                    i = s.index('=')
                    params[unquote(s[:i])] = unquote(s[i+1:])

            if path.startswith('webseed/'):
                return self.handle_webseed(path, params)
            else:
                return self.send_error(404, 'File not found')

        except ValueError, e:
            return self.send_error(400, 'you sent me garbage - ' + str(e))
        except Exception, e:
            return self.send_error(500, 'some unhandled error - ' + str(e))

    def handle_webseed(self, path, params):
        if not self.server.webseed_semaphore.acquire(False):
            return self.send_error(503)
        try:
            file_ = path[len('webseed/'):]
            name = file_.split('/')[0]
            files = flatten(map(itemgetter('files'),
                                self.server.tracker.allowed.itervalues()))

            if not os.path.join(*file_.split('/')) in files:
                logger().warning("request for invalid file '%s'" % file_)
                return self.send_error(404, 'you sent me garbage - file '
                                            'does not exist')
            p = os.path.join(self.server.tracker.files_dir, file_)
            ct = self.guess_type(self.path)
            response_code = 200
            headers = dict()
            if 'Range' in self.headers:
                range_ = self.headers['Range']
                if not range_.startswith('bytes='):
                    raise ValueError('invalid range given')

                start, end = [int(x) for x in range_[6:].split('-')]
                if start < 0 or end < 0 or start > end:
                    raise ValueError('invalid range given')

                end = min(end, start + self.server.tracker.piecesizes[name] - 1)
                length = end - start + 1
                # Prepare response
                response_code = 206
                headers.update({
                    'Content-Range': 'bytes %i-%i/*' % (start, end)
                })
            else:
                if os.stat(p).st_size > self.server.tracker.piecesizes[name]:
                    raise ValueError('requested size exceeds piece length')
                length = None
            # Send response
            headers.update({
                'Content-Type': ct,
            })

            logger().info("%f: sending %d bytes" % (time.time(), length))
            return self.send_file(response_code, headers, p, start, length)
        finally:
            self.server.webseed_semaphore.release()

class Webseeder(HTTPServer):
    def __init__(self, port, tracker, max_handler_threads=10, maxupload=-1):
        logger().info('initializing webseed server ...')

        self.doneevent = Event()
        self.tracker = tracker
        self.port = port

        HTTPServer.__init__(self, ('', port), WebseedRequestHandler)

        if maxupload > 0:
            maxupload *= 1024
        elif maxupload == 0:
            maxupload = -1
        self.upload_limiter = RateLimiter(maxupload)

        self.webseed_semaphore = Semaphore(max_handler_threads)
        self.request_queue = Queue()
        for i in xrange(max_handler_threads + 1):
            Thread(target = self._process_request_thread).start()

    def _process_request_thread(self):
        while not self.doneevent.isSet():
            try:
                request, client_address = self.request_queue.get()
                if request is None and self.doneevent.isSet():
                    # Shutdown next thread
                    # XXX do we really need that one? same line in do_QUIT
                    self.request_queue.put((None, None))
                    return

                self.finish_request(request, client_address)
            except socket.error, e:
                logger().exception(e)
            finally:
                if request is not None:
                    self.close_request(request)

    def process_request(self, request, client_address):
        self.request_queue.put((request, client_address))

    def serve_forever(self):
        logger().info('webseed server starts serving ...')
        while not self.doneevent.isSet():
            self.handle_request()

    def shutdown(self):
        logger().info('shutting down webseed server')

        # just force the server to handle a request
        self.doneevent.set()
        try:
            conn = HTTPConnection('localhost:%d' % self.port)
            conn.request("QUIT", "/")
            conn.getresponse()
        except socket.error, e:
            logger().exception(e)

        logger().info('webseed server shut down done')
